/*
 * dwxout.c - Output data according to DWARF information
 *
 * Written 2002 by Werner Almesberger, Caltech Netlab FAST project
 */


#include <inttypes.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "dwxtype.h"
#include "dwarfhelp.h"
#include "dwxout.h"


#define INDENT_STEP 4


/* ----- Helper functions -------------------------------------------------- */


static void indent(FILE *file,int level)
{
    fprintf(file,"%*s",level*INDENT_STEP,"");
}


static unsigned long read_size(const void *p,int size)
{
    switch (size) {
	case 1:
	    return *(uint8_t *) p;
	case 2:
	    return read_16(p);
	case 4:
	    return read_32(p);
	case -1:
	    fprintf(stderr,"integer size %d not supported\n",size);
	    exit(1);
	default:
	    abort();
    }
}


void dwx_output_hexdump(FILE *file,const void *data,int length,int indent)
{
    const unsigned char *p = data;
    int width,i,j;

    width = (79-8)/4;
    if (width > 8) width &= ~7;
    if (width < 4) width = 16; /* don't make it ridiculously small */
    for (i = 0; i < length; i += 16) {
	fprintf(file,"%*s",indent,"");
	fprintf(file,"%4x: ",i);
	for (j = i; j < i+width && j < length; j++) {
	    if (!(j & 7)) fputc(' ',file);
	    fprintf(file,"%02x ",p[j]);
	}
	while (j < i+width) {
	    if (!(j & 7)) fputc(' ',file);
	    fprintf(file,"   ");
	    j++;
	}
	for (j = i; j < i+width && j < length; j++) {
	    if (!(j & 7)) fputc(' ',file);
	    fprintf(file,"%c",*p < ' ' || *p > '~' ? '.' : *p);
	    p++;
	}
	fputc('\n',file);
    }
}


/* ----- Guesstimate size of type ------------------------------------------ */


int dwx_sizeof(const struct dwx_type *t)
{
    switch (t->kind) {
	const struct dwx_struct_member *m;
	int size,max;

	case dwx_uint:
	case dwx_sint:
	case dwx_float:
	    assert(t->size != -1);
	    return t->size;
	case dwx_enum:
	    return t->size == -1 ? 4 : t->size;
	case dwx_const:
	case dwx_volatile:
	    if (t->size != -1) return t->size;
	    return dwx_sizeof(t->u.modifier.type);
	case dwx_ptr:
	    return 4;
	case dwx_array:
	    if (t->size != -1) return t->size;
	    if (t->u.array.stride_size == -1)
		return t->u.array.elements*t->u.array.stride_size;
	    return t->u.array.elements*dwx_sizeof(t->u.array.type);
	case dwx_struct:
	    if (t->size != -1) return t->size;
	    size = 0;
	    for (m = t->u._struct.members; m; m = m->next)
		size += dwx_sizeof(m->type);
	    return size;
	case dwx_union:
	    if (t->size != -1) return t->size;
	    max = 0;
	    for (m = t->u._struct.members; m; m = m->next) {
		size = dwx_sizeof(m->type);
		if (size > max) max = size;
	    }
	    return max;
	case dwx_typedef:
	    return dwx_sizeof(t->u._typedef.type);
	default:
	    abort();
    }
}


/* ----- Type output ------------------------------------------------------- */


static void output_enum_members_type(FILE *file,const struct dwx_enum_member *m,
  int level)
{
    while (m) {
	indent(file,level);
	fprintf(file,"%s = %d\n",m->name,m->value);
	m = m->next;
    }
}


static void do_output_type(FILE *file,const struct dwx_type *t,int level,
  int terse);


static void output_struct_members_type(FILE *file,
  const struct dwx_struct_member *m,int level)
{
    while (m) {
	indent(file,level);
	fprintf(file,".%s <@%lu>\n",m->name,m->offset);
	do_output_type(file,m->type,level+1,0);
	m = m->next;
    }
}


static void do_output_type(FILE *file,const struct dwx_type *t,int level,
  int terse)
{
    indent(file,level);
    if (!t) {
	fprintf(file,"void\n");
	return;
    }
    switch (t->kind) {
	case dwx_uint:
	    switch (t->size) {
		case 1:
		    fprintf(file,"unsigned char\n");
		    break;
		case 2:
		    fprintf(file,"uint16_t\n");
		    break;
		case 4:
		    fprintf(file,"uint32_t\n");
		    break;
		case -1:
		    fprintf(file,"unsigned int <%d>\n",t->size);
		    break;
		default:
		    abort();
	    }
	    break;
	case dwx_sint:
	    switch (t->size) {
		case 1:
		    fprintf(file,"char\n");
		    break;
		case 2:
		    fprintf(file,"int16_t\n");
		    break;
		case 4:
		    fprintf(file,"int32_t\n");
		    break;
		case -1:
		    fprintf(file,"int <%d>\n",t->size);
		    break;
		default:
		    abort();
	    }
	    break;
	case dwx_float:
	    fprintf(file,"float <%d>\n",t->size);
	    break;
	case dwx_enum:
	    fprintf(file,"enum %s <%d>\n",t->name ? t->name : "",t->size);
	    output_enum_members_type(file,t->u._enum.members,level+1);
	    break;
	case dwx_const:
	    fprintf(file,"const <%d>\n",t->size);
	    do_output_type(file,t->u.modifier.type,level+1,terse);
	    break;
	case dwx_ptr:
	    fprintf(file,"* <%d>\n",t->size);
	    do_output_type(file,t->u.modifier.type,level+1,1);
	    break;
	case dwx_volatile:
	    fprintf(file,"volatile <%d>\n",t->size);
	    do_output_type(file,t->u.modifier.type,level+1,terse);
	    break;
	case dwx_fn:
	    fprintf(file,"fn() <%d>\n",t->size);
	    do_output_type(file,t->u.fn.type,level+1,1);
	    break;
	case dwx_array:
	    fprintf(file,"[%u] <%d, *%d>\n",t->u.array.elements,t->size,
	      t->u.array.stride_size);
	    do_output_type(file,t->u.array.type,level+1,terse);
	    break;
	case dwx_struct:
	    fprintf(file,"struct %s <%d>\n",t->name ? t->name : "",t->size);
	    if (!terse)
		output_struct_members_type(file,t->u._struct.members,level+1);
	    break;
	case dwx_union:
	    fprintf(file,"union %s <%d>\n",t->name ? t->name : "",t->size);
	    if (!terse)
		output_struct_members_type(file,t->u._struct.members,level+1);
	    break;
	case dwx_typedef:
	    fprintf(file,"typedef %s <%d>\n",t->name,t->size);
	    do_output_type(file,t->u._typedef.type,level+1,terse);
	    break;
	default:
	    abort();
    }
}


void dwx_output_type(FILE *file,const struct dwx_type *type)
{
    do_output_type(file,type,0,0);
}


/* ----- Data output ------------------------------------------------------- */


static void do_output_data(FILE *file,const struct dwx_type *t,
  const void *data,unsigned long offset,int level);


static void output_enum_members_data(FILE *file,
  const struct dwx_enum_member *m,unsigned long value)
{
    while (m) {
	if (m->value == value) {
	    fprintf(file,"%s\n",m->name);
	    return;
	}
	m = m->next;
    }
    fprintf(file,"%lu\n",value);
}


static void output_struct_members_data(FILE *file,
  const struct dwx_struct_member *m,const void *data,unsigned long offset,
  int level)
{
    while (m) {
	indent(file,level);
	fprintf(file,"%s = ",m->name);
	do_output_data(file,m->type,data,offset+m->offset,level);
	m = m->next;
    }
}


static void do_output_data(FILE *file,const struct dwx_type *t,
  const void *data,unsigned long offset,int level)
{
    const void *p;
    int i;

    p = data+offset;
    switch (t->kind) {
	case dwx_uint:
	    {
		unsigned long tmp;

		tmp = read_size(p,t->size);
		fprintf(file,"%lu (0x%lx)\n",tmp,tmp);
	    }
	    break;
	case dwx_sint:
	    {
		long tmp;

		switch (t->size) {
		    case 1:
			tmp = *(int8_t *) p;
			break;
		    case 2:
			tmp = read_16(p);
			if (tmp & 0x8000) tmp |= ~0 & ~0xffff;
			break;
		    case 4:
			tmp = read_32(p);
			if (tmp & 0x80000000) tmp |= ~0 & ~0xffffffff;
			break;
		    default:
			fprintf(stderr,"integer size %d not supported\n",
			  t->size);
			exit(1);
		}
		fprintf(file,"%ld (0x%lx)\n",tmp,tmp);
	    }
	    break;
	case dwx_float:
	    fprintf(stderr,"we don't support floats yet, sorry\n");
	    exit(1);
	case dwx_enum:
	    output_enum_members_data(file,t->u._enum.members,
	      read_size(p,t->size == -1 ? 4 : t->size));
	    break;
	case dwx_ptr:
	    fprintf(file,"0x%lx\n",(unsigned long) read_32(p));
	    break;
	case dwx_const:
	case dwx_volatile:
	    do_output_data(file,t->u.modifier.type,data,offset,level);
	    break;
	case dwx_array:
	    {
		const struct dwx_type *type;

		type = t->u.array.type;
		if (type->kind == dwx_uint && type->size == 1 &&
		  t->u.array.stride_size <= 1) { /* stride size -1 or 1 */
		    fputc('\n',file);
		    dwx_output_hexdump(file,p,t->u.array.elements,
		      (level+1)*INDENT_STEP);
		    break;
		}
		if (type->kind == dwx_sint && type->size == 1 &&
		  t->u.array.stride_size <= 1) {
		    int last;

		    fputc('"',file);
		    last = -1;
			/* using -1 here makes sure we don't print the "no \0"
			   warning in the case of an empty array, i.e. x[0] */
		    for (i = 0; i < t->u.array.elements; i++)
			if (*(const char *) (p+i)) last = i;
		    for (i = 0; i < last; i++) {
			unsigned char c = *(const unsigned char *) (p+i);

			if (c == '\\' || c == '"') fprintf(file,"\\%c",c);
			else if (c >= ' ' && c <= '~') fputc(c,file);
			else fprintf(file,"\\%03o",c);
		    }
		    if (last == t->u.array.elements) fprintf(file," (no \\0)");
		    fprintf(file,"\"\n");
		    break;
		}
		fprintf(file,"{\n");
		for (i = 0; i < t->u.array.elements; i++) {
		    indent(file,level+1);
		    fprintf(file,"[%d] = ",i);
		    do_output_data(file,t->u.array.type,data,offset,level+1);
		    if (t->u.array.stride_size == -1)
			offset += dwx_sizeof(type);
		    else offset += t->u.array.stride_size;
		}
		indent(file,level);
		fprintf(file,"}\n");
	    }
	    break;
	case dwx_struct:
	    fprintf(file,"{\n");
	    output_struct_members_data(file,t->u._struct.members,data,offset,
	      level+1);
	    indent(file,level);
	    fprintf(file,"}\n");
	    break;
	case dwx_union:
	    fprintf(file,"{\n");
	    output_struct_members_data(file,t->u._struct.members,data,offset,
	      level+1);
	    indent(file,level);
	    fprintf(file,"}\n");
	    break;
	case dwx_typedef:
	    do_output_data(file,t->u._typedef.type,data,offset,level);
	    break;
	default:
	    abort();
    }
}


void dwx_output_data(FILE *file,const struct dwx_type *type,const void *data)
{
    do_output_data(file,type,data,0,0);
}


/* ----- Variable lookup --------------------------------------------------- */


static const struct dwx_type *lookup_var(const struct dwx_type *t,va_list ap,
  unsigned long *offset)
{
    while (1) {
	switch (t->kind) {
	    const struct dwx_struct_member *m;
	    const char *name;
	    int num;

	    case dwx_uint:
	    case dwx_sint:
	    case dwx_enum:
	    case dwx_ptr:
		return t;
	    case dwx_const:
	    case dwx_volatile:
		t = t->u.modifier.type;
		break;
	    case dwx_typedef:
		t = t->u._typedef.type;
		break;
	    case dwx_array:
		num = va_arg(ap,int);
		*offset += num*(t->u.array.stride_size != -1 ?
		  t->u.array.stride_size : dwx_sizeof(t->u.array.type));
		t = t->u.array.type;
		break;
	    case dwx_struct:
	    case dwx_union:
		name = va_arg(ap,const char *);
		for (m = t->u._struct.members; m; m = m->next)
		     if (!strcmp(m->name,name)) break;
		if (!m) {
		     fprintf(stderr,"no member \"%s\"\n",name);
		     exit(1);
		}
		*offset += m->offset;
		t = m->type;
		break;
	    default:
		abort();
	}
    }
}


unsigned long dwx_lookup_unsigned(const struct dwx_type *t,
  const void *data,...)
{
    unsigned long offset = 0;
    va_list ap;

    va_start(ap,data);
    t = lookup_var(t,ap,&offset);
    va_end(ap);
    return read_size(data+offset,t->size);
}


long dwx_lookup_signed(const struct dwx_type *t,const void *data,...)
{
    unsigned long offset = 0;
    va_list ap;

    va_start(ap,data);
    t = lookup_var(t,ap,&offset);
    va_end(ap);
    return read_size(data+offset,t->size); /* @@@ need to sign-extend ... */
}
